import { alias, int, mysqlTable, text, timestamp, varchar } from 'drizzle-orm/mysql-core';
import type { MySql2Database } from 'drizzle-orm/mysql2';
import { and, eq } from 'drizzle-orm';
import { reportStatus } from './reportStatus.ts';
import { generateRandomString } from '@util/random.ts';
import { team } from '@db/schema/team.ts';
import { BASE_PATH } from 'astro:env/server';
import { strikeReason } from '@db/schema/strikeReason.ts';
import { strike } from '@db/schema/strike.ts';

type Database = MySql2Database<{ report: typeof report }>;

export const report = mysqlTable('report', {
  id: int('id').primaryKey().autoincrement(),
  reason: varchar('reason', { length: 255 }).notNull(),
  body: text('body'),
  urlHash: varchar('url_hash', { length: 255 }).notNull(),
  createdAt: timestamp('created_at', { mode: 'date' }),
  reporterTeamId: int('reporter_team_id').references(() => team.id),
  reportedTeamId: int('reported_team_id').references(() => team.id)
});

export type AddReportReq = {
  reason: string;
  body: string | null;
  createdAt?: Date | null;
  reporterTeamId?: number;
  reportedTeamId?: number | null;
};

export type SubmitReportReq = {
  urlHash: string;
  reason: string;
  body: string;
};

export type GetReportsReq = {
  reporter?: string | null;
  reported?: string | null;
};

export type GetReportById = {
  id: number;
};

export type GetReportByUrlHash = {
  urlHash: string;
};

export async function addReport(db: Database, values: AddReportReq) {
  const urlHash = generateRandomString(16);

  const r = await db
    .insert(report)
    .values({
      reason: values.reason,
      body: values.body,
      urlHash: urlHash,
      createdAt: values.createdAt,
      reporterTeamId: values.reporterTeamId,
      reportedTeamId: values.reportedTeamId
    })
    .$returningId();

  return Object.assign(r[0], { url: `${BASE_PATH}/report/${urlHash}` });
}

export async function submitReport(db: Database, values: SubmitReportReq) {
  return db
    .update(report)
    .set({
      reason: values.reason,
      body: values.body,
      createdAt: new Date()
    })
    .where(eq(report.urlHash, values.urlHash));
}

export async function getReports(db: Database, values: GetReportsReq) {
  const reporterTeam = alias(team, 'reporter');
  const reportedTeam = alias(team, 'reported');

  let reporterIdSubquery;
  if (values.reporter != null) {
    reporterIdSubquery = db
      .select({ id: reporterTeam.id })
      .from(reporterTeam)
      .where(eq(reporterTeam.name, values.reporter))
      .as('reporter_id_subquery');
  }
  let reportedIdSubquery;
  if (values.reported != null) {
    reportedIdSubquery = db
      .select({ id: reportedTeam.id })
      .from(reportedTeam)
      .where(eq(reportedTeam.name, values.reported))
      .as('reported_id_subquery');
  }

  return db
    .select({
      id: report.id,
      reason: report.reason,
      body: report.body,
      createdAt: report.createdAt,
      reporter: {
        id: reporterTeam.id,
        name: reporterTeam.name
      },
      reported: {
        id: reportedTeam.id,
        name: reportedTeam.name
      },
      status: {
        status: reportStatus.status,
        notice: reportStatus.notice,
        statement: reportStatus.statement
      },
      strike: {
        strikeReasonId: strikeReason.id
      }
    })
    .from(report)
    .innerJoin(reporterTeam, eq(report.reporterTeamId, reporterTeam.id))
    .leftJoin(reportedTeam, eq(report.reportedTeamId, reportedTeam.id))
    .leftJoin(reportStatus, eq(report.id, reportStatus.reportId))
    .leftJoin(strike, eq(report.id, strike.reportId))
    .leftJoin(strikeReason, eq(strike.strikeReasonId, strikeReason.id))
    .where(
      and(
        values.reporter != null ? eq(report.reporterTeamId, reporterIdSubquery!.id) : undefined,
        values.reported != null ? eq(report.reportedTeamId, reportedIdSubquery!.id) : undefined
      )
    );
}

export async function getReportById(db: Database, values: GetReportById) {
  const reporterTeam = alias(team, 'reporter');
  const reportedTeam = alias(team, 'reported');

  const reports = await db
    .select({
      id: report.id,
      reason: report.reason,
      body: report.body,
      createdAt: report.createdAt,
      reporter: {
        id: reporterTeam.id,
        name: reporterTeam.name
      },
      reported: {
        id: reportedTeam.id,
        name: reportedTeam.name
      },
      status: {
        status: reportStatus.status,
        notice: reportStatus.notice,
        statement: reportStatus.statement
      }
    })
    .from(report)
    .innerJoin(reporterTeam, eq(report.reporterTeamId, reporterTeam.id))
    .leftJoin(reportedTeam, eq(report.reportedTeamId, reportedTeam.id))
    .leftJoin(reportStatus, eq(report.id, reportStatus.reportId))
    .where(eq(report.id, values.id));

  return reports[0] ?? null;
}

export async function getReportByUrlHash(db: Database, values: GetReportByUrlHash) {
  const reporterTeam = alias(team, 'reporter');
  const reportedTeam = alias(team, 'reported');

  const reports = await db
    .select({
      id: report.id,
      reason: report.reason,
      body: report.body,
      createdAt: report.createdAt,
      urlHash: report.urlHash,
      reporter: {
        id: reporterTeam.id,
        name: reporterTeam.name
      },
      reported: {
        id: reportedTeam.id,
        name: reportedTeam.name
      },
      status: {
        status: reportStatus.status,
        notice: reportStatus.notice,
        statement: reportStatus.statement
      }
    })
    .from(report)
    .innerJoin(reporterTeam, eq(report.reporterTeamId, reporterTeam.id))
    .leftJoin(reportedTeam, eq(report.reportedTeamId, reportedTeam.id))
    .leftJoin(reportStatus, eq(report.id, reportStatus.reportId))
    .where(eq(report.urlHash, values.urlHash));

  return reports[0] ?? null;
}
